@@ -5,17 +5,16 @@ from selenium.common.exceptions import WebDriverException
 from selenium.webdriver.common.keys import Keys
 from selenium.webdriver.support.ui import WebDriverWait
 from selenium.webdriver.support import expected_conditions
+from selenium.webdriver.common.keys import Keys
 import webbrowser
-from pathlib import Path
-from cleverdict import CleverDict
 import pyperclip
-import inspect
-import json
+from pathlib import Path
 import keyring
-from .clevergui import start_gui, button_menu, text_input
-from .cleverutils import to_json, timer
-from .cleverwebutils import Login_to
+from .clevergui import *
+from .cleverweb import *
+from .cleverutils import *
 import threading
+import difflib
 
 class CleverSession(CleverDict):
     """
@@ -34,20 +33,54 @@ class CleverSession(CleverDict):
     index = CleverDict()
     choices = {"https://github.com/login": "Github",
                "https://twitter.com": "Twitter",
-               "https://www.satchelone.com/login": "SatchelOne"}
+               "https://www.satchelone.com/login": "SatchelOne",
+               "https://www.hackerrank.com": "HackerRank",
+               "192.168.0.1": "TP-Link"}
     keyring_config_root = keyring.util.platform_.config_root()
     keyring_data_root = keyring.util.platform_.data_root()
 
     def __init__(self, **kwargs):
+        options, kwargs = self.get_options_from_kwargs(**kwargs)
         super().__init__(**kwargs)
+        self.start_time = get_time()
         start_gui(redirect=kwargs.get("redirect"))
         self.check_and_prompt("url")
         if self.get("url"):
             self.account = CleverSession.choices[self.url]
+            self.login_url = self.url
             self.get_username()
         if not self.get("dirpath"):
-            self.dirpath = Path().cwd()
-        self.max_browsers = 5
+            self.dirpath_str = str(Path().cwd())
+        self.max_browsers = 5 if self.account == "Github" else 1
+        if kwargs.get("echo") is True:
+            setattr(CleverSession, "save", CleverSession.echo_on)
+        if kwargs.get("echo") is False:
+            setattr(CleverSession, "save", CleverSession.echo_off)
+        self.browser = kwargs.get("browser")
+        if not self.browser:
+            self.browser = webdriver.Chrome(options=disable_logging(**kwargs))
+        self.browser.implicitly_wait(kwargs.get("wait") or 3)
+        self.add_current_browser()
+
+    @property
+    def dirpath(self):
+        """
+        This @property returns a pathlib.Path version of .dirpath_str.
+        Needed in order to allow JSON serialisation without constant conversions
+        since pathlib.Path is not JSON serialisable.
+        """
+        return Path(self.dirpath_str)
+
+    def get_options_from_kwargs(self, **kwargs):
+        """ Separate actionable options from general data in kwargs."""
+        options = {}
+        for key, default_value in {"echo": True, "_break": False, "redirect": False}.items():
+            if isinstance(kwargs.get(key), bool):
+                options[key] = kwargs.get(key)
+                del kwargs[key]
+            else:
+                options[key] = default_value
+        return options, kwargs
 
     def get_username(self):
         """
@@ -103,6 +136,12 @@ class CleverSession(CleverDict):
         """
         keyring.delete_password(CleverSession.choices[self.url], self.username)
 
+    def add_current_browser(self):
+        """Appends the current (login) browser to self.browsers"""
+        if not hasattr(self, "browsers"):
+            self.browsers = []
+        self.browsers += [self.browser]
+
     @timer
     def login_with_webbrowsers(self, browsers=None):
         """
@@ -111,37 +150,83 @@ class CleverSession(CleverDict):
         wait : Seconds for selenium to implicitly_wait
         browsers : int > number of browsers to run concurrently
         """
-        self.check_and_prompt("url", "username", "password")
-        if not hasattr(self, "browsers"):
-            self.setattr_direct("browsers", [])
-        dispatch = {"github.com": Login_to.github,
-                    "twitter.com": Login_to.twitter,
-                    "satchelone.com": Login_to.satchelone}
-        browserThreads = []
-        if browsers is None:
-            browsers = self.get("max_browsers") or 1
-        # Adjust if main browser is already be running:
-        browsers = browsers - len(self.browsers)
-        for n in  range(browsers):
-            for website, func in dispatch.items():
-                if website in self.url:
-                    browserThread = threading.Thread(target=func, args=[self])
-                    browserThreads.append(browserThread)
-                    browserThread.start()
-                    break
-        for browserThread in browserThreads:
-                browserThread.join()
+        try:
+            self.check_and_prompt("url", "username", "password")
+            if not hasattr(self, "browsers"):
+                self.setattr_direct("browsers", [])
+            dispatch = {"github.com": Login_to.github,
+                        "twitter.com": Login_to.twitter,
+                        "satchelone.com": Login_to.satchelone,
+                        "hackerrank.com": Login_to.hackerrank,
+                        "192.168.0.1": Login_to.tplink}
+            browserThreads = []
+            if browsers is None:
+                browsers = self.get("max_browsers") or 1
+            # Adjust if main browser is already be running:
+            browsers = browsers - len(self.browsers)
+            for n in  range(browsers):
+                for website, func in dispatch.items():
+                    if website in self.url:
+                        browserThread = threading.Thread(target=func, args=[self])
+                        browserThreads.append(browserThread)
+                        browserThread.start()
+                        break
+            for browserThread in browserThreads:
+                    browserThread.join()
+        except WebDriverException:
+            raise WebDriverException("Check chromdriver is in your PATH or you're running this code from a directory with chromedriver.exe in it")
 
-    def save(self, name, value):
-        """ Generic auto-save confirmation applied CleverDict """
+    def echo_on(self, name, value):
+        """
+        Generic confirmation applied CleverDict auto-save with:
+
+        setattr(CleverSession, "save", CleverSession.echo_on)
+        """
         if "password" not in str(name).lower() and name not in vars(self):
             # vars(self) includes CleverDict keys created with setattr_direct()
             # i.e. not intended to be readily accessible as data attributes
             print(f" ⓘ  {name} = {value} {type(value)}")
 
+    def echo_off(self, name, value):
+        """
+        Disable CleverSession autosave confirmations with:
+
+        setattr(CleverSession, "save", CleverSession.echo_off)
+        """
+        pass
+
     def start(self):
         """ Shortcut/Alias for starting a webbrowser session and logging in """
-        self.login_with_webbrowser()
+        self.login_with_webbrowsers()
+
+    def listen(self, save=False):
+        """
+        Compares web page source code (.page_source) with a previous
+        version/page of the same browser session (.source), calculates any
+        differences using difflib, and saves the changed lines (only) to
+        .source_diff and to file (if save=True).
+
+        self.browser is a selenium webdriver instance.
+        """
+        if not hasattr(self, "source"):
+            self.source = ""
+            self.source_diff = []
+            self.source_index = 1
+        source = self.browser.page_source.splitlines()
+        if source != self.source:
+            print("\nⓘ  Page source code has changed.  Saving to .source_diff")
+            d = difflib.Differ()
+            diff = d.compare(source, self.source)
+            diff = '\n'.join([x for x in diff if x[0] in "+-?"])
+            if self.source:
+                self.source_diff += [diff]
+                if save:
+                    filename = "source" + str(self.source_index).zfill(3)
+                    with open(filename, "w") as file:
+                        file.write(diff)
+                        print(f"\nⓘ  Saved page source to {filename}")
+                        self.source_index += 1
+            self.source = source
 
 
 # setattr(CleverSession, "login_with_webbrowsers", login_with_webbrowsers)
